package org.processmining.earthmoversstochasticconformancechecking.distancematrix;

import java.util.concurrent.atomic.AtomicInteger;

import org.apache.commons.lang3.ArrayUtils;
import org.processmining.earthmoversstochasticconformancechecking.stochasticlanguage.StochasticLanguage;
import org.processmining.earthmoversstochasticconformancechecking.stochasticlanguage.StochasticTraceIterator;
import org.processmining.framework.plugin.ProMCanceller;

public abstract class DistanceMatrixAbstract implements DistanceMatrix {

	private double[] distances;

	private int languageBsize;

	public void init(StochasticLanguage languageA, final StochasticLanguage languageB, final ProMCanceller canceller)
			throws InterruptedException {
		languageBsize = languageB.size();
		distances = new double[languageA.size() * languageB.size() + 1];

		final AtomicInteger al = new AtomicInteger(0);
		final StochasticTraceIterator itA = languageA.iterator();

		Thread[] threads = new Thread[Math.max(1, Runtime.getRuntime().availableProcessors() - 1)];
		for (int thread = 0; thread < threads.length; thread++) {
			Runnable runnable = new Runnable() {
				public void run() {
					thread(itA, languageB, al, canceller);
				}
			};
			threads[thread] = new Thread(runnable, "distances thread " + thread);
			threads[thread].start();
		}

		for (int thread = 0; thread < threads.length; thread++) {
			threads[thread].join();
			threads[thread] = null;
		}
	}

	public void thread(StochasticTraceIterator itA, StochasticLanguage languageB, AtomicInteger al,
			ProMCanceller canceller) {
		while (true) {

			//get a trace from the log
			int[] traceA;
			int l;
			synchronized (itA) {
				l = al.getAndIncrement();
				if (!itA.hasNext()) {
					return;
				}
				traceA = itA.nextIntegerTrace();
			}

			if (canceller.isCancelled()) {
				return;
			}

			StochasticTraceIterator itB = languageB.iterator();
			for (int m = 0; m < languageB.size(); m++) {
				//StochasticPath pathB = itB.next();
				//String[] traceB = StochasticPath2Trace.convert(pathB);
				int[] traceB = itB.nextIntegerTrace();
				double distance = getDistance(traceA, traceB);
				set(l, m, distance);

				if (canceller.isCancelled()) {
					return;
				}
			}
		}
	}

	/**
	 * A method that computes the actual distance. Should be thread-safe.
	 * 
	 * @param traceA
	 * @param traceB
	 * @return
	 */
	protected abstract double getDistance(int[] traceA, int[] traceB);

	public double[] getDistances() {
		return distances;
	}

	public double getDistance(int l, int m) {
		return distances[l * languageBsize + m + 1];
	}

	private void set(int l, int m, double distance) {
		distances[l * languageBsize + m + 1] = distance;
	}

	public DistanceMatrixAbstract clone() {
		DistanceMatrixAbstract result;
		try {
			result = (DistanceMatrixAbstract) super.clone();
			result.distances = ArrayUtils.clone(distances);
			result.languageBsize = languageBsize;
			return result;
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
			return null;
		}
	}

}