package org.processmining.earthmoversstochasticconformancechecking.stochasticlanguage.model;

import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.commons.lang3.ArrayUtils;
import org.processmining.earthmoversstochasticconformancechecking.helperclasses.EfficientStochasticPetriNetSemantics;
import org.processmining.earthmoversstochasticconformancechecking.helperclasses.EfficientStochasticPetriNetSemanticsImpl;
import org.processmining.earthmoversstochasticconformancechecking.helperclasses.PrefixProbabilityMarking;
import org.processmining.earthmoversstochasticconformancechecking.helperclasses.QueueCombination;
import org.processmining.earthmoversstochasticconformancechecking.helperclasses.TransitionMap;
import org.processmining.earthmoversstochasticconformancechecking.parameters.TerminationStrategy;
import org.processmining.earthmoversstochasticconformancechecking.stochasticlanguage.Activity2IndexKey;
import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.models.graphbased.directed.petrinet.StochasticNet;
import org.processmining.models.semantics.IllegalTransitionException;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.InductiveMiner.Pair;

import com.google.common.util.concurrent.AtomicDouble;

/**
 * Use multithreading and byte storage
 * 
 * @author sander
 *
 */
public class StochasticPetriNet2StochasticPathLanguage {

	/**
	 * Assumption: the deadlock markings are equal to the final markings.
	 * 
	 * @param net
	 * @param initialMarking
	 * @param terminationStrategy
	 * @param canceller
	 * @return
	 * @throws IllegalTransitionException
	 * @throws InterruptedException
	 */
	public static Pair<StochasticLanguageModel, TransitionMap> convert(StochasticNet net, Marking initialMarking,
			TerminationStrategy terminationStrategy, Activity2IndexKey activityKey, ProMCanceller canceller)
			throws InterruptedException {

		//initialise
		EfficientStochasticPetriNetSemanticsImpl semantics = new EfficientStochasticPetriNetSemanticsImpl(net,
				initialMarking);
		TransitionMap transitionMap = new TransitionMap(net, semantics);
		StochasticTransition2IndexKey transitionKey = new StochasticTransition2IndexKey(semantics, activityKey);
		StochasticPathLanguageImpl language = new StochasticPathLanguageImpl(transitionKey, activityKey);

		//set to work
		walk(language, semantics, transitionMap, initialMarking, terminationStrategy, canceller);

		if (canceller.isCancelled()) {
			return null;
		}

		return Pair.of((StochasticLanguageModel) language, transitionMap);
	}

	public static double walk(final StochasticPathLanguageImpl language,
			final EfficientStochasticPetriNetSemantics semantics, final TransitionMap transitionMap,
			Marking initialMarking, TerminationStrategy terminationStrategy, final ProMCanceller canceller)
			throws InterruptedException {
		//initialise queues
		final ConcurrentLinkedQueue<byte[]> queue = new ConcurrentLinkedQueue<>();
		final AtomicInteger queueSize = new AtomicInteger(1);

		//add the first step to the queues
		queue.add(PrefixProbabilityMarking.pack(new int[0], 1.0, semantics.convert(initialMarking)));

		final AtomicDouble massCovered = new AtomicDouble(0);
		final TerminationStrategy terminationStrategy2 = terminationStrategy.clone();
		terminationStrategy2.initialise();

		Thread[] threads = new Thread[Math.max(1, Runtime.getRuntime().availableProcessors() - 1)];
		for (int thread = 0; thread < threads.length; thread++) {
			Runnable runnable = new Runnable() {
				public void run() {
					try {
						thread(semantics.clone(), transitionMap, queue, queueSize, massCovered, terminationStrategy2,
								language, canceller);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			};
			threads[thread] = new Thread(runnable, "stochastic language thread " + thread);
			threads[thread].start();
		}

		for (int thread = 0; thread < threads.length; thread++) {
			threads[thread].join();
			threads[thread] = null;
		}

		if (canceller.isCancelled()) {
			return -Double.MAX_VALUE;
		}

		return massCovered.get();
	}

	public static double getTotalMass(EfficientStochasticPetriNetSemantics semantics, int[] enabledTransitions) {
		double sum = 0;
		for (int i : enabledTransitions) {
			sum += semantics.getTransitionWeight(i);
		}
		return sum;
	}

	public static void thread(EfficientStochasticPetriNetSemantics semantics, TransitionMap transitionMap,
			ConcurrentLinkedQueue<byte[]> globalQueue, AtomicInteger globalQueueSize, AtomicDouble massCovered,
			TerminationStrategy terminationStrategy, StochasticPathLanguageImpl language, ProMCanceller canceller)
			throws InterruptedException {
		final int markingLength = semantics.getState().length;

		final QueueCombination queue = new QueueCombination(globalQueue, globalQueueSize);
		//ConcurrentLinkedQueue<byte[]> queue = globalQueue;

		byte[] prefixProbabilityMarking;
		while (true) {

			if (canceller.isCancelled()) {
				return;
			}

			prefixProbabilityMarking = queue.poll();
			if (prefixProbabilityMarking == null) {
				if (terminationStrategy.isTerminated(massCovered.get())) {
					return;
				} else {
					Thread.sleep(10);
					continue;
				}
			}

			int[] prefix = PrefixProbabilityMarking.getPrefix(prefixProbabilityMarking, markingLength);
			double prefixProbability = PrefixProbabilityMarking.getProbability(prefixProbabilityMarking);
			byte[] marking = PrefixProbabilityMarking.getMarking(prefixProbabilityMarking, markingLength);

			semantics.setState(marking);
			int[] enabledTransitions = semantics.getEnabledTransitions();
			double totalTransitionsMass = getTotalMass(semantics, enabledTransitions);

			if (enabledTransitions.length == 0) {
				//we are in a deadlock state. by assumption, that's a final state.

				//accept the current trace and add it to the log
				synchronized (language) {
					if (terminationStrategy.isTerminated(massCovered.get())) {
						return;
					}
					language.add(prefix2path(semantics, transitionMap, prefix), prefixProbability);
					if (terminationStrategy.isTerminated(massCovered.addAndGet(prefixProbability))) {
						return;
					}
				}

				//System.out.println(prefix + " accepted, total " + massCovered);

				//				System.out.println(prefixProbability + " " + Arrays.toString(prefix) + " total " + massCovered);

			} else {
				//we are not in a deadlock state; continue
				for (int transition : enabledTransitions) {

					semantics.setState(marking);
					semantics.executeTransition(transition);
					byte[] newMarking = semantics.getState();

					int[] newPrefix = ArrayUtils.add(prefix, transition);

					//compute the new probability
					double newProbability = prefixProbability * semantics.getTransitionWeight(transition)
							/ totalTransitionsMass;

					queue.add(PrefixProbabilityMarking.pack(newPrefix, newProbability, newMarking));
				}
			}
		}
	}

	private static int[] prefix2path(EfficientStochasticPetriNetSemantics semantics, TransitionMap transitionMap,
			int[] prefix) {
		int[] result = new int[prefix.length];
		System.arraycopy(prefix, 0, result, 0, prefix.length);
		return result;
	}
}
